/*
 * @(#)gl_texture.java 0.4 01/03/15
 *
 * jGL 3-D graphics library for Java
 * Copyright (c) 1999-2001 Robin Bing-Yu Chen (robin@is.s.u-tokyo.ac.jp)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version. the GNU Lesser
 * General Public License should be included with this distribution
 * in the file LICENSE.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

package jgl.context.attrib;

// import java.lang.System;

import jgl.GL;
// import jgl.context.gl_list_item;
import jgl.context.gl_util;
import jgl.context.gl_image;
import jgl.context.attrib.texture.*;

/**
 * gl_texture is the texturing class of jGL 2.4.
 *
 * @version 	0.4, 15 Mar 2001
 * @author 	Robin Bing-Yu Chen
 */

public class gl_texture {

/*
    public static final int TEXTURE_1D = 1;
    public static final int TEXTURE_2D = 2;
    public static final int TEXTURE_3D = 4;

    public static final int Q_BIT = 1;
    public static final int R_BIT = 2;
    public static final int S_BIT = 4;
    public static final int T_BIT = 8;
*/

    /** GL_TEXTURE_x: True if x-D texturing enabled (x is 1D, 2D, or 3D) */
//    public int Enable = 0;	/* Bitwise-OR of TEXTURE_[123]D values */

    /** For texture x-D */
    public gl_texture_obj Current1D = new gl_texture_obj ();
    public gl_texture_obj Current2D = new gl_texture_obj ();
    public gl_texture_obj Current3D = new gl_texture_obj ();
    public gl_texture_obj CurrentObj;

    /** GL_TEXTURE_ENV_MODE: Texture application function */
    public int EnvMode = GL.GL_MODULATE;

    /** GL_TEXTURE_ENV_COLOR: Texture environment color */
    public float EnvColor [] = {0, 0, 0, 0};

    /** GL_TEXTURE_GEN_x: Texgen enabled (x is S, T, R, Q) */
//    public int GenEnabled = 0;		/* Bitwise-OR of [QRST]_BIT values */

    /** For texgen x */
    public gl_texture_gen CurrentS = new gl_texture_gen ();
    public gl_texture_gen CurrentT = new gl_texture_gen ();
    public gl_texture_gen CurrentR = new gl_texture_gen ();
    public gl_texture_gen CurrentQ = new gl_texture_gen ();

    public boolean check_tex () {
	if (Current3D.Enable) { CurrentObj = Current3D; return true; }
	if (Current2D.Enable) { CurrentObj = Current2D; return true; }
	if (Current1D.Enable) { CurrentObj = Current1D; return true; }
	return false;
    }

    public int is_tex_enabled () {
	int TexEnabled = 0;
	if (Current1D.Enable) TexEnabled = TexEnabled | 1;
	if (Current2D.Enable) TexEnabled = TexEnabled | 2;
	if (Current3D.Enable) TexEnabled = TexEnabled | 4;
	return TexEnabled;
    }

    public boolean is_tex_enabled (int cap) {
	switch (cap) {
	    case GL.GL_TEXTURE_1D: return (Current1D.Enable);
	    case GL.GL_TEXTURE_2D: return (Current2D.Enable);
	    case GL.GL_TEXTURE_3D: return (Current3D.Enable);
	} return false;
    }
    
    public boolean tex_enable (int state) {
	if ((state & 1)!=0) Current1D.Enable=true; else Current1D.Enable=false;
	if ((state & 2)!=0) Current2D.Enable=true; else Current2D.Enable=false;
	if ((state & 4)!=0) Current3D.Enable=true; else Current3D.Enable=false;
	return check_tex ();
    }

    public boolean tex_enable (int cap, boolean state) {
	switch (cap) {
	    case GL.GL_TEXTURE_1D: Current1D.Enable = state; break;
	    case GL.GL_TEXTURE_2D: Current2D.Enable = state; break;
	    case GL.GL_TEXTURE_3D: Current3D.Enable = state; break;
	} return check_tex ();
    }
    
    public int is_tex_gen_enabled () {
    	int GenEnabled = 0;
    	if (CurrentQ.Enable) GenEnabled = GenEnabled | 1;
    	if (CurrentR.Enable) GenEnabled = GenEnabled | 2;
    	if (CurrentS.Enable) GenEnabled = GenEnabled | 4;
    	if (CurrentT.Enable) GenEnabled = GenEnabled | 8;
    	return GenEnabled;
    }
    
    public boolean is_tex_gen_enabled (int cap) {
    	switch (cap) {
    	    case GL.GL_TEXTURE_GEN_Q: return (CurrentQ.Enable);
    	    case GL.GL_TEXTURE_GEN_R: return (CurrentR.Enable);
    	    case GL.GL_TEXTURE_GEN_S: return (CurrentS.Enable);
    	    case GL.GL_TEXTURE_GEN_T: return (CurrentT.Enable);
	} return false;
    }
    
    public void tex_gen_enable (int state) {
	if ((state & 1)!=0) CurrentQ.Enable=true; else CurrentQ.Enable=false;
	if ((state & 2)!=0) CurrentR.Enable=true; else CurrentR.Enable=false;
	if ((state & 4)!=0) CurrentS.Enable=true; else CurrentS.Enable=false;
	if ((state & 8)!=0) CurrentT.Enable=true; else CurrentT.Enable=false;
    }
    
    public void tex_gen_enable (int cap, boolean state) {
    	switch (cap) {
    	    case GL.GL_TEXTURE_GEN_Q: CurrentQ.Enable = state; break;
    	    case GL.GL_TEXTURE_GEN_R: CurrentR.Enable = state; break;
    	    case GL.GL_TEXTURE_GEN_S: CurrentS.Enable = state; break;
    	    case GL.GL_TEXTURE_GEN_T: CurrentT.Enable = state; break;
    	}
    }

    public void set_tex (int target, int pname, float params []) {
	switch (target) {
	    case GL.GL_TEXTURE_1D: Current1D.set_tex (pname, params); break;
	    case GL.GL_TEXTURE_2D: Current2D.set_tex (pname, params); break;
	    case GL.GL_TEXTURE_3D: Current3D.set_tex (pname, params); break;
	}
    }

    public void set_tex_image (int target, int level, int border, gl_image img) {
	switch (target) {
	    case GL.GL_TEXTURE_1D: Current1D.set_tex_image(level,border,img);break;
	    case GL.GL_TEXTURE_2D: Current2D.set_tex_image(level,border,img);break;
	    case GL.GL_TEXTURE_3D: Current3D.set_tex_image(level,border,img);break;
	}
    }

    public void set_tex_gen_i (int coord, int param) {
	switch (coord) {
	    case GL.GL_Q: CurrentQ.Mode = param; break;
	    case GL.GL_R: CurrentR.Mode = param; break;
	    case GL.GL_S: CurrentS.Mode = param; break;
	    case GL.GL_T: CurrentT.Mode = param; break;
	}
    }

    public void set_tex_gen_f (int coord, int pname, float params []) {
	switch (coord) {
	    case GL.GL_Q: CurrentQ.set_tex_gen (pname, params); break;
	    case GL.GL_R: CurrentR.set_tex_gen (pname, params); break;
	    case GL.GL_S: CurrentS.set_tex_gen (pname, params); break;
	    case GL.GL_T: CurrentT.set_tex_gen (pname, params); break;
	}
    }

    /****** will delete the following ******/

    private int tex_map_nearest (float s, float t, int l) {
    	gl_image img = CurrentObj.Image [l].Image;
	int w = img.Width;
	int h = img.Height;
    	int i, j;

	if (CurrentObj.WrapS == GL.GL_REPEAT) {
	    i = (int)(s * w);
	    i = i & (w - 1);
	} else { // CurrentObj.WrapS == GL.GL_CLAMP
	    if (s < 0) { i = 0;
	    } else { if (s >= 1) { i = w - 1;
	    } else { i = (int)(s * w); } }
	}
	if (CurrentObj.WrapT == GL.GL_REPEAT) {
	    j = (int)(t * h);
	    j = j & (h - 1);
	} else { // CurrentObj.WrapT == GL.GL_CLAMP
	    if (t < 0) { j = 0;
	    } else { if (t >= 1) { j = h - 1;
	    } else { j = (int)(t * h); } }
	}
	return (0xff000000 | (img.ImageData [i][j][0][0] & 0x000000ff) << 16 |
	    		     (img.ImageData [i][j][0][1] & 0x000000ff) <<  8 |
			     (img.ImageData [i][j][0][2] & 0x000000ff));
    }

    private int tex_map_linear (float s, float t, int l) {
    	gl_image img = CurrentObj.Image [l].Image;
	int w = img.Width;
	int h = img.Height;
	float u, v;
	float a, b;
	int i0, j0, i1, j1;
	float w00, w01, w10, w11;
	int rgb [] = new int [3];

	u = s * w;
	if (CurrentObj.WrapS == GL.GL_REPEAT) {
	    i0 = ((int)Math.floor (u - (float)0.5)) & (w - 1);
	    i1 = (i0 + 1) & (w - 1);
	} else { // CurrentObj.WrapS == GL.GL_CLAMP
	    i0 = (int)Math.floor (u - (float)0.5);
	    i1 = i0 + 1;
	}

	v = t * h;
	if (CurrentObj.WrapT == GL.GL_REPEAT) {
	    j0 = ((int)Math.floor (v - (float)0.5)) & (h - 1);
	    j1 = (j0 + 1) & (h - 1);
	} else { // CurrentObj.WrapT == GL.GL_CLAMP
	    j0 = (int)Math.floor (v - (float)0.5);
	    j1 = j0 + 1;
	}

	a = gl_util.frac (u - (float)0.5);
	b = gl_util.frac (v - (float)0.5);

	w00 = (1 - a) * (1 - b); w01 = a  * (1 - b);
	w10 = (1 - a) *      b ; w11 = a  *      b ;

	rgb[0]=(int)(((float)(img.ImageData[i0][j0][0][0]&0x000000ff))*w00+
		     ((float)(img.ImageData[i0][j1][0][0]&0x000000ff))*w01+
		     ((float)(img.ImageData[i1][j0][0][0]&0x000000ff))*w10+
		     ((float)(img.ImageData[i1][j1][0][0]&0x000000ff))*w11);
	rgb[1]=(int)(((float)(img.ImageData[i0][j0][0][1]&0x000000ff))*w00+
		     ((float)(img.ImageData[i0][j1][0][1]&0x000000ff))*w01+
		     ((float)(img.ImageData[i1][j0][0][1]&0x000000ff))*w10+
		     ((float)(img.ImageData[i1][j1][0][1]&0x000000ff))*w11);
	rgb[2]=(int)(((float)(img.ImageData[i0][j0][0][2]&0x000000ff))*w00+
		     ((float)(img.ImageData[i0][j1][0][2]&0x000000ff))*w01+
		     ((float)(img.ImageData[i1][j0][0][2]&0x000000ff))*w10+
		     ((float)(img.ImageData[i1][j1][0][2]&0x000000ff))*w11);

//	if (rgb[0] == 0 && rgb[1] == 0 && rgb[2] == 0)
//	System.out.println ("ERROR "+s+" "+t);
//	return (0xff000000 | rgb [0] << 16 | rgb [1] << 8 | rgb [2]);
	return gl_util.RGBtoI(rgb[0], rgb[1], rgb[2]);
    }

    private int tex_map (float s, float t) {
	if (CurrentObj.MagFilter == GL.GL_NEAREST) {
	    return tex_map_nearest (s, t, 0);
	} else { // CurrentObj.MagFilter == GL.GL_LINEAR
	    return tex_map_linear (s, t, 0);
	}
    }

    private int tex_mid_nearest_nearest (float s, float t, float l) {
    	int level;
	if (l <= 0.5) { level = 0;
	} else {
	    level = (int)(l + 0.499999);
	    if (level > CurrentObj.Image [0].WidthLog2) {
	    	level = CurrentObj.Image [0].WidthLog2;
	    }
	}
	return tex_map_nearest (s, t, level);
    }

    private int tex_mid_linear_nearest (float s, float t, float l) {
    	int level;
	if (l <= 0.5) { level = 0;
	} else {
	    level = (int)(l + 0.499999);
	    if (level > CurrentObj.Image [0].WidthLog2) {
	    	level = CurrentObj.Image [0].WidthLog2;
	    }
	}
	return tex_map_linear (s, t, level);
    }

    private int tex_mid_nearest_linear (float s, float t, float l) {
    	int max = CurrentObj.Image [0].MaxLog2;
    	if (l >= max) {
	    return tex_map_nearest (s, t, max);
	} else {
	    float f = gl_util.frac (l);
	    int level = (int)(l + 1.0);
	    int rgb [][] = new int [2][3];
	    if (level < 1)   { level = 1; }
	    if (level > max) { level = max; }

	    rgb [0][0] = tex_map_nearest (s, t, level - 1);
	    rgb [1][0] = tex_map_nearest (s, t, level);

	    rgb [0][2] = rgb [0][0] & 0x00ff0000;
	    rgb [0][1] = rgb [0][0] & 0x0000ff00;
	    rgb [0][0] = rgb [0][0] & 0x000000ff;
	    rgb [1][2] = rgb [1][0] & 0x00ff0000;
	    rgb [1][1] = rgb [1][0] & 0x0000ff00;
	    rgb [1][0] = rgb [1][0] & 0x000000ff;

	    rgb [0][0] = (int)(((float)rgb[1][0])*((float)1.0-f)+
	    		       ((float)rgb[0][0])*f);
	    rgb [0][1] = (int)(((float)rgb[1][1])*((float)1.0-f)+
	    		       ((float)rgb[0][1])*f);
	    rgb [0][2] = (int)(((float)rgb[1][2])*((float)1.0-f)+
	    		       ((float)rgb[0][2])*f);

//	    return (0xff000000 | rgb[0][0] << 16 | rgb[0][1] << 8 | rgb[0][2]);
	    return gl_util.RGBtoI(rgb[0][0], rgb[0][1], rgb[0][2]);
	}
    }

    private int tex_mid_linear_linear (float s, float t, float l) {
    	int max = CurrentObj.Image [0].MaxLog2;
    	if (l >= max) {
	    return tex_map_linear (s, t, max);
	} else {
	    float f = gl_util.frac (l);
	    int level = (int)(l + 1.0);
	    int rgb [][] = new int [2][3];
	    if (level < 1)   { level = 1; }
	    if (level > max) { level = max; }

	    rgb [0][0] = tex_map_linear (s, t, level - 1);
	    rgb [1][0] = tex_map_linear (s, t, level);

	    rgb [0][2] = rgb [0][0] & 0x00ff0000;
	    rgb [0][1] = rgb [0][0] & 0x0000ff00;
	    rgb [0][0] = rgb [0][0] & 0x000000ff;
	    rgb [1][2] = rgb [1][0] & 0x00ff0000;
	    rgb [1][1] = rgb [1][0] & 0x0000ff00;
	    rgb [1][0] = rgb [1][0] & 0x000000ff;

	    rgb [0][0] = (int)(((float)rgb[1][0])*((float)1.0-f)+
	    		       ((float)rgb[0][0])*f);
	    rgb [0][1] = (int)(((float)rgb[1][1])*((float)1.0-f)+
	    		       ((float)rgb[0][1])*f);
	    rgb [0][2] = (int)(((float)rgb[1][2])*((float)1.0-f)+
	    		       ((float)rgb[0][2])*f);

//	    return (0xff000000 | rgb[0][0] << 16 | rgb[0][1] << 8 | rgb[0][2]);
	    return gl_util.RGBtoI(rgb[0][0], rgb[0][1], rgb[0][2]);
	}
    }

    private int tex_mid (float s, float t, float l) {
    	float c;

   	if (CurrentObj.MagFilter == GL.GL_LINEAR &&
	   (CurrentObj.MinFilter == GL.GL_NEAREST_MIPMAP_NEAREST ||
	    CurrentObj.MinFilter == GL.GL_LINEAR_MIPMAP_NEAREST)){ c=(float)0.5;
	} else { c = 0; }

	if (l > c) {
	    switch (CurrentObj.MinFilter) {
	        case GL.GL_NEAREST_MIPMAP_NEAREST:
	            return tex_mid_nearest_nearest (s, t, l);
	        case GL.GL_LINEAR_MIPMAP_NEAREST:
	            return tex_mid_linear_nearest (s, t, l);
	        case GL.GL_NEAREST_MIPMAP_LINEAR:
	            return tex_mid_nearest_linear (s, t, l);
	        case GL.GL_LINEAR_MIPMAP_LINEAR:
	            return tex_mid_linear_linear (s, t, l);
	        default:
	            return 0;
	    }
	} else {
	    return tex_map (s, t);
	}
    }

    private float lambda (float w,float dsdx,float dsdy,float dtdx,float dtdy) {
	float dudx, dudy, dvdx, dvdy, r1, r2, rho;

	if (dsdx < 0) { dudx = -dsdx; } else { dudx = dsdx; }
	if (dsdy < 0) { dudy = -dsdy; } else { dudy = dsdy; }
	if (dtdx < 0) { dvdx = -dtdx; } else { dvdx = dtdx; }
	if (dtdy < 0) { dvdy = -dtdy; } else { dvdy = dtdy; }

	r1 = Math.max (dudx, dudy) * CurrentObj.Image [0].Width;
	r2 = Math.max (dvdx, dvdy) * CurrentObj.Image [0].Height;
	rho = Math.max (r1, r2) / w;

	if (rho < 0) { rho = -rho; }

	if (rho == 0) { return 0; } 
	
	return ((float)(Math.log(rho)*1.442695)); // 1.442695=1/log2
    }

    public int tex_vertex (float s, float t, float w, float dsdx, float dsdy,
						      float dtdx, float dtdy) {
   	if (CurrentObj.MinFilter == GL.GL_NEAREST ||
	    CurrentObj.MinFilter == GL.GL_LINEAR) {
	    return tex_map (s, t);
	} else {
	    return tex_mid (s, t, lambda (w, dsdx, dsdy, dtdx, dtdy));
	}
    }

    public int tex_vertex (float s, float t) {
	return tex_map (s, t);
    }

    public float [] get_tex_gen_coord (float obj[], float eye[], float nor[]) {
	float TexCoord [] = new float [4];
	if (CurrentS.Enable) TexCoord[0] = CurrentS.get_tex_gen_coord(obj, eye, nor);
			else TexCoord[0] = 0;
	if (CurrentT.Enable) TexCoord[1] = CurrentT.get_tex_gen_coord(obj, eye, nor);
			else TexCoord[1] = 0;
	if (CurrentR.Enable) TexCoord[2] = CurrentR.get_tex_gen_coord(obj, eye);
			else TexCoord[2] = 0;
	if (CurrentQ.Enable) TexCoord[3] = CurrentQ.get_tex_gen_coord(obj, eye);
			else TexCoord[3] = 1;
	return TexCoord;
    }

/*
    public void push_attrib (gl_list_item AttribItem) {
	AttribItem.BoolPtr  = new boolean [4];
	AttribItem.IntPtr   = new int     [1];
	AttribItem.FloatPtr = new float   [4];

    public gl_texture_obj Current1D = new gl_texture_obj ();
    public gl_texture_obj Current2D = new gl_texture_obj ();
    public gl_texture_obj Current3D = new gl_texture_obj ();
    public int EnvMode = GL.GL_MODULATE;
    public float EnvColor [] = {0, 0, 0, 0};
    public gl_texture_gen CurrentS = new gl_texture_gen ();
    public gl_texture_gen CurrentT = new gl_texture_gen ();
    public gl_texture_gen CurrentR = new gl_texture_gen ();
    public gl_texture_gen CurrentQ = new gl_texture_gen ();

	AttribItem.IntPtr  [0] = ClearIndex;
	System.arraycopy(ClearColor, 0, AttribItem.FloatPtr, 0, 4);
	AttribItem.IntPtr  [1] = IntClearColor;
	AttribItem.IntPtr  [2] = IndexMask;
	AttribItem.IntPtr  [3] = ColorMask;
	AttribItem.BoolPtr [0] = AlphaEnable;
	AttribItem.IntPtr  [4] = AlphaFunc;
	AttribItem.IntPtr  [5] = AlphaRef;
	AttribItem.BoolPtr [1] = BlendEnable;
	AttribItem.IntPtr  [6] = BlendSrc;
	AttribItem.IntPtr  [7] = BlendDst;
	AttribItem.BoolPtr [2] = LogicOPEnable;
	AttribItem.IntPtr  [8] = LogicOPMode;
	AttribItem.BoolPtr [3] = DitherEnable;
    }

    public void pop_attrib (gl_list_item AttribItem) {
	ClearIndex    = AttribItem.IntPtr  [0];
	System.arraycopy(AttribItem.FloatPtr, 0, ClearColor, 0, 4);
	IntClearColor = AttribItem.IntPtr  [1];
	IndexMask     = AttribItem.IntPtr  [2];
	ColorMask     = AttribItem.IntPtr  [3];
	AlphaEnable   = AttribItem.BoolPtr [0];
	AlphaFunc     = AttribItem.IntPtr  [4];
	AlphaRef      = AttribItem.IntPtr  [5];
	BlendEnable   = AttribItem.BoolPtr [1];
	BlendSrc      = AttribItem.IntPtr  [6];
	BlendDst      = AttribItem.IntPtr  [7];
	LogicOPEnable = AttribItem.BoolPtr [2];
	LogicOPMode   = AttribItem.IntPtr  [8];
	DitherEnable  = AttribItem.BoolPtr [3];
    }
*/

}
